/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
//


#include "ns3/log.h"
#include "ns3/address.h"
#include "ns3/node.h"
#include "ns3/nstime.h"
#include "ns3/data-rate.h"
#include "ns3/random-variable.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/tag.h"
#include "ns3/uinteger.h"
#include "ns3/string.h"
#include "ns3/trace-source-accessor.h"
#include "pbc.h"
#include "ns3/udp-socket-factory.h"
#include "ns3/inet-socket-address.h"
#include "ns3/double.h"
#include "ns3/node-list.h"
#include "ns3/integer.h"
#include "ns3/wifi-net-device.h"
#include "ns3/yans-wifi-phy.h"
#include "ns3/config.h"
#include "ns3/boolean.h"

NS_LOG_COMPONENT_DEFINE ("PbcApplication");

using namespace std;

namespace ns3 {


NS_OBJECT_ENSURE_REGISTERED (PbcApplication);

TypeId
PbcApplication::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::PbcApplication")
    .SetParent<Application> ()
    .AddConstructor<PbcApplication> ()
    .AddAttribute ("Interval",
                   "The time interval between two successive broadcast transmissions.",
                   TimeValue (Seconds(0.1)),
                   MakeTimeAccessor (&PbcApplication::m_interval),
                   MakeTimeChecker())
    .AddAttribute ("PacketSize",
                   "Packet size in Bytes.",
                   IntegerValue (400),
                   MakeIntegerAccessor (&PbcApplication::m_packetSize),
                   MakeIntegerChecker<uint32_t> ())
    .AddAttribute ("RemoteAddress",
                   "The address of the destination (either a Unicast or a Broadcast address)",
                   AddressValue (),
                   MakeAddressAccessor (&PbcApplication::m_peer),
                   MakeAddressChecker ())
    .AddAttribute ("Protocol",
                   "The type of protocol to use.",
                   TypeIdValue (UdpSocketFactory::GetTypeId ()),
                   MakeTypeIdAccessor (&PbcApplication::m_protocol),
                   MakeTypeIdChecker ())
    .AddAttribute ("ActivateReceive",
                   "Listen for other packets",
                   BooleanValue (false),
                   MakeBooleanAccessor (&PbcApplication::m_recvact),
                   MakeBooleanChecker ())
    .AddAttribute ("ActivateCbrTrace",
                   "Observe Channel Busy Rate.",
                   BooleanValue (false),
                   MakeBooleanAccessor (&PbcApplication::m_pcbract),
                   MakeBooleanChecker ())
    .AddTraceSource ("CbrTrace", 
                   "channel busy rate",
                   MakeTraceSourceAccessor (&PbcApplication::m_cbrTrace))
    .AddTraceSource ("SendTrace", 
                   "packet send",
                   MakeTraceSourceAccessor (&PbcApplication::m_sendTrace)) 
    .AddTraceSource ("RecvTrace", 
                   "packet received",
                   MakeTraceSourceAccessor (&PbcApplication::m_recvTrace))                                    
    ;
  return tid;
}


PbcApplication::PbcApplication()
 : m_socket(0),
 m_sendcount(0),
 m_busyTimes(0),
 m_idleTimes(0)
   
   
{
  NS_LOG_FUNCTION_NOARGS ();
}

PbcApplication::~PbcApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
}

void
PbcApplication::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_socket = 0;

    // chain up
  Application::DoDispose ();

}

void
PbcApplication::StartApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
  
  
  // init position and direction
  m_pos = GetNode()->GetObject<MobilityModel>()->GetPosition();
  m_nodeId = GetNode()->GetId(); 
  

    // Create the socket if not already
  if (!m_socket)
    {
      m_socket = Socket::CreateSocket (GetNode(), m_protocol);
      if (m_recvact)
        m_socket->SetRecvCallback (MakeCallback (&PbcApplication::ReceivePacket, this));
      m_socket->Bind ();
      m_socket->Connect (m_peer);
      
    }
    
    // Ensure no pending event
  CancelEvents ();



  if (m_pcbract) {
    std::ostringstream oss;
    oss.str ("");
    oss << "/NodeList/" << GetNode()->GetId() << "/DeviceList/*";
    std::string devicepath = oss.str ();
    Config::Connect (devicepath + "/Phy/State/State", MakeCallback (&PbcApplication::PhyStateTrace, this));
  }


  if (m_interval.GetSeconds () > 0) {
    // Start application with jitter
    UniformVariable jitterGenerator(0, m_interval.GetSeconds ());
    m_jitter = jitterGenerator.GetValue();
    Time m_startJitter = Seconds(m_jitter);
    m_sendEvent = Simulator::Schedule(m_startJitter, &PbcApplication::SendPacket, this);
  }

  //m_cbrEvent = Simulator::Schedule(Seconds(5), &PbcApplication::UpdateCBR, this);
  
}

void
PbcApplication::StopApplication()
{
  NS_LOG_FUNCTION_NOARGS ();

  if (m_pcbract)
    m_cbrTrace (CalculateDistance (m_pos, Vector()), (double)m_busyTimes.GetNanoSeconds()/(m_idleTimes.GetNanoSeconds()+m_busyTimes.GetNanoSeconds()));
  
  CancelEvents ();
  if(m_socket != 0)
    {
      m_socket->Close ();
    }
  else
    {
      NS_LOG_WARN("PbcApplication found null socket to close in StopApplication");
    }
}

void
PbcApplication::CancelEvents ()
{
  NS_LOG_FUNCTION_NOARGS ();
  Simulator::Cancel(m_sendEvent);
}


void
PbcApplication::ReceivePacket (Ptr<Socket> socket)
{
  
  NS_LOG_FUNCTION (this << socket);
  while (m_socket->GetRxAvailable () > 0)
  {
    Address from;
    Ptr<Packet> p = m_socket->RecvFrom (0xffffffff, 0, from);
    
    if (p->GetSize () == m_packetSize+1) {
      m_recvTrace();
    }
  }    

}



void
PbcApplication::SendPacket()
{
  NS_LOG_FUNCTION_NOARGS ();
//  NS_LOG_INFO ("PbcApplication::SendPacket() at " << Simulator::Now().GetSeconds() << "sec");
  NS_ASSERT (m_sendEvent.IsExpired ());
  Ptr<Packet> packet;
  Time nextTime(m_interval);

  if (GetNode()->GetId() == 1) {
    m_sendTrace ();
    packet = Create<Packet> (m_packetSize+1);
    // std::cout << "sent at " << Simulator::Now().GetSeconds() << std::endl;
  } else {
    packet = Create<Packet> (m_packetSize);
  }
    // Send packet and trace the event
  int rc = m_socket->Send (packet);
  
  if (rc < 0) {
    std::cout << "###################socket error" << m_socket->GetErrno()<<std::endl;
  }
  
  
  if (m_sendcount++ % 10 == 0) {
    // rejitter
    UniformVariable jitterGenerator(0, m_interval.GetSeconds ());
    double newJitter = jitterGenerator.GetValue();
    double diff = newJitter - m_jitter;
    m_jitter = newJitter;
    nextTime = Seconds(m_interval.GetSeconds() - diff);
  }
  
  m_sendEvent = Simulator::Schedule(nextTime, &PbcApplication::SendPacket, this);

}

void
PbcApplication::PhyStateTrace (std::string context, Time start, Time duration, enum WifiPhy::State state)
{
//std::cout << Simulator::Now ().GetSeconds() << "#####phy trace" << std::endl;
  if (state == WifiPhy::IDLE) {
    m_idleTimes += duration;
  } else {
    m_busyTimes += duration;
  }
}


void 
PbcApplication::UpdateCBR (void)
{
//  m_cbrTrace (CalculateDistance (m_pos, Vector()), (double)m_busyTimes.GetNanoSeconds()/(m_idleTimes.GetNanoSeconds()+m_busyTimes.GetNanoSeconds()));
  
//  m_cbrEvent = Simulator::Schedule(Seconds(5), &PbcApplication::UpdateCBR, this);

}

} // Namespace ns3
