/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * 
 *
 */
 
 
#include "ns3/log.h"
#include "manhattan-topology-helper.h"
#include "ns3/mobility-helper.h"
#include "ns3/integer.h"
#include <sstream>
#include "ns3/double.h"

#include "ns3/packet-socket-address.h"
#include "ns3/traffic-light.h"
#include "ns3/string.h"
#include "ns3/boolean.h"


NS_LOG_COMPONENT_DEFINE ("ManhattanTopologyHelper");

namespace ns3 {


ManhattanTopologyHelper::ManhattanTopologyHelper ()
{

  m_mainDir = 0;
  m_laneWidth = 4;
  m_sourceLength = 80;
  m_setNoConflict = 0;
  m_makeTrafficLights = 0;
  
  speedvar = NormalVariable  (50.0/3.6, (3/3.6)*(3/3.6), 9/3.6);
  thvar = NormalVariable (1.5, 0.2*0.2, 0.5);
  accvar = NormalVariable (1.5, 0.2*0.2, 0.5);
  decvar = NormalVariable (2.5, 0.2*0.2, 0.5);
  ivtvar = ExponentialVariable (20);
  ivtmainvar = ExponentialVariable (10);
}
ManhattanTopologyHelper::~ManhattanTopologyHelper ()
{
}


/*
 * Connects a road with two intersections and sets length, dir, etc.
 * the road will be set 'laneWidth' meters to the right from the intersections
 * if only one intersection is given, direction must be set (for source and sink roads)
 */
void
ManhattanTopologyHelper::connectRoad (Ptr<Road> road, Ptr<Intersection> from, 
      Ptr<Intersection> to, Vector direction = Vector(0,0,0))
{
  Vector frompos;
  Vector topos;
  
  NS_ASSERT(from != NULL || to != NULL);
  
  if (from != NULL) {
    frompos = from->m_pos;
    from->m_roadsFrom.push_back (road);
  } else {
    // source road: set startpos in reverse direction (vdiff)
    frompos = IDMUtils::vdiff(to->m_pos, IDMUtils::vmult(direction, m_sourceLength));
  }
  if (to != NULL) {
    topos = to->m_pos;
    to->AddApproachRoad (road);
  } else {
    // sink road: set topos in reverse direction (vplus)
    topos = IDMUtils::vplus(from->m_pos, IDMUtils::vmult(direction, m_sourceLength)); 
  }
  

  Vector diff = IDMUtils::vdiff (topos, frompos);
  double length = IDMUtils::vlength(diff);
  
  // calc real start pos from intersection+lanewidth
  // rotate direction 90° to right and norm length to 1
  Vector dirorthogonal = IDMUtils::vnorm(Vector(diff.y, -1*diff.x, 0));
  road->m_startPos = IDMUtils::vplus(frompos, IDMUtils::vmult(dirorthogonal, m_laneWidth*0.5));
  
  // source road
  if (from == NULL) {
    m_sourceRoads.push_front (road);
  } else {
    road->m_start = from;
  }
  
  road->m_end = to; // to can be NULL (end of road)
  road->m_dir = IDMUtils::vnorm(diff);
  road->m_length = length;
}

/*
 * connects all intersections between the two intersections indicated by starti/j
 * and endi/j by calling connectRoad on every single pair of adjacent intersections.
 * also creates source and sink roads using direction.
 * if bothDir is true, calls itself for opposite direction.
 */
void
ManhattanTopologyHelper::makeSinglePath (int starti, int startj, int endi, int endj,
    Vector direction, bool bothDir)
{
  NS_ASSERT((starti == endi) || (startj == endj));  // only straight path
  Ptr<Intersection> prevInt;
  
  int i = starti;
  int j = startj;
  
  // source road    
  connectRoad (Road::roads[m_roadC], NULL, iget(i,j), direction);
  
  // mark the last two added roads at this intersection as non conflicting
  if (m_setNoConflict && !bothDir) {
    iget(i,j)->AddNoConflict();
  }
  prevInt = iget(i,j);
  m_roadC++;
  
  // roads between intersections
  while ((i != endi) || (j != endj)) {
    if (i > endi)
      i--;
    if (i < endi)
      i++;
    if (j > endj)
      j--;
    if (j < endj)
      j++;
      
    connectRoad (Road::roads[m_roadC], prevInt, iget(i,j));
    
    // mark the last two added roads at this intersection as non conflicting
    if (m_setNoConflict && !bothDir) {
      iget(i,j)->AddNoConflict();
    }
    m_roadC++;
    prevInt = iget(i,j);
  }
  
  // sink road
  connectRoad (Road::roads[m_roadC], prevInt, NULL, direction);  
  m_roadC++;
  
  if (bothDir)
    makeSinglePath (endi, endj, starti, startj, IDMUtils::vmult(direction, -1.0), false);
    
}

/*
 * creates a manhattan grid topology with blocksV*blocksH intersections and according
 * block sizes. also creates source and sink roads from/to outside intersections
 * also creates vehicles sources on source roads by calling makeVehicleSources
 * on every source road.
 */
void
ManhattanTopologyHelper::makeSimpleTopo (int blocksV, int blocksH, double blockSizeV, 
      double blockSizeH, double ibd, bool suburban, Time endTime)
{
  NS_ASSERT (blocksV > 0 && blocksH > 0);
  NS_LOG_DEBUG ("making topology");
  
  int nIntersections = blocksV*blocksH;
  int nSources = 2*blocksV + 2*blocksH;
  Vector direction;

  Ptr<Intersection> intersection;
  int i, j;
  
  Road::roads.resize(4*nIntersections + nSources);
  
  m_blocksH = blocksH;
  m_blocksV = blocksV;
  m_blockSizeV = blockSizeV;
  m_blockSizeH = blockSizeH;
  
  
  
  // create all intersections (connect and configure later)
  for (i = 0; i < blocksV; i++) {
    for (j = 0; j < blocksH; j++) {
      intersection = iget(i, j);
      intersection->m_pos = Vector((j+1)*blockSizeH, (i+1)*blockSizeV, 0);
      intersection->m_id = i*blocksH+j;
      intersection->m_stopDist = m_laneWidth;
      //std::cout << "# "<< iget(i, j)->m_pos << std::endl;
    } 
  }
  
  // create all roads
  for (i = 0; i < 4*nIntersections + nSources; i++) {
    Road::roads[i] = CreateObject<Road>();
    Road::roads[i]->m_id = i;
  }
  
  
  // connect roads and intersections
  
  m_roadC = 0;
  
  // up down, down up
  ///////
  for (i = 0; i < blocksH; i++) {
    direction = Vector(0,1,0);
    makeSinglePath (0, i, blocksV-1, i, direction, true);
  }
  
  if (m_mainDir == 1) {
    for (i = 0; i < m_roadC; i++) {
      Road::roads[i]->m_prio = IDM_HIGH_PRIO;
    }
  }  
  
  
  // left to right / right to left
  ///////
  for (i = 0; i < blocksV; i++) {
    direction = Vector(1,0,0);
    makeSinglePath (i, 0, i, blocksH-1, direction, true);
  }  
    


  // as road network is connected now, let roads get their neighbors
  std::vector<Ptr<Road> >::iterator it;
  for (it = Road::roads.begin() ; it != Road::roads.end(); ++it) {
    (*it)->CheckNextIntersection ();
    //std::cout << "# roads #" << (*it)->m_startPos << " " << (*it)->m_length << std::endl;
  }
  
  NS_LOG_DEBUG ("roads connected, making nodes");





  NodeContainer nc;
  std::list<Ptr<Road> >::iterator roadit;
  for (roadit=m_sourceRoads.begin(); roadit != m_sourceRoads.end(); ++roadit) {
    makeVehicleSources (*roadit, endTime);
  }
    
  if (m_makeTrafficLights) {
    std::vector<Ptr<Intersection> >::iterator intit;
    for (intit = m_intersections.begin(); intit != m_intersections.end(); ++intit) {
      makeTrafficLights(*intit, endTime);
    }
  }

  NS_LOG_DEBUG ("maketopo finished");
}

void
ManhattanTopologyHelper::makeTrafficLights (Ptr<Intersection> curint, Time endTime) {
  PacketSocketAddress pbcSocket;
  pbcSocket.SetAllDevices();
  pbcSocket.SetPhysicalAddress(Mac48Address::GetBroadcast());
  pbcSocket.SetProtocol(1);
  
  ObjectFactory factory;
  factory.SetTypeId (TrafficLight::GetTypeId ());
  factory.Set ("Protocol", StringValue ("ns3::PacketSocketFactory"));
  factory.Set ("RemoteAddress", AddressValue (Address(pbcSocket)));
  factory.Set ("StartTime", TimeValue(Seconds(0.0)));
  factory.Set ("StopTime", TimeValue(endTime));
  factory.Set ("Interval", TimeValue(Seconds(0.1)));
  factory.Set ("PhaseLength", TimeValue(Seconds(30)));
  factory.Set ("Synchronized", BooleanValue(m_stlsync));
  
  Ptr<TrafficLight> app = factory.Create<TrafficLight> ();
  app->m_int = curint;
  Ptr<Node> tmpNode = CreateObject<Node> ();
  tmpNode->AddApplication (app);
  
  MobilityHelper mobility;
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel",
                                  "Position", VectorValue(curint->m_pos)
                                  );  // does not work, set in tl-app!
  mobility.Install (tmpNode);

  m_nc.Add (tmpNode);
}


void
ManhattanTopologyHelper::makeVehicleSources (Ptr<Road> source, Time endTime)
{
  double curTime, add, diff;
  double speed, acc, dec, timeHeadway;
  curTime = 0;
  diff = 0;
  while (true) {
  
    timeHeadway = thvar.GetValue();
    speed = speedvar.GetValue();
    acc = accvar.GetValue ();
    dec = decvar.GetValue ();
    
    
    
    // prevent vehicles from being generated too close to each other
    // keep exponential distribution but ensure min_timeHeadway
    if (source->m_prio == IDM_HIGH_PRIO) {
      add = ivtmainvar.GetValue () - diff;
    } else {
      add = ivtvar.GetValue () - diff;
    }
    if (curTime == 0 || add >= timeHeadway) {
      curTime += add;
      diff = 0;
    } else {
      curTime += timeHeadway;
      diff = timeHeadway - add;
    }
    
    //std::cout << "curtime = " << curTime << std::endl;
    if (curTime + 1 > endTime.GetSeconds ())
      break;
      
    Ptr<Node> tmpNode = CreateObject<Node> ();
    m_nc.Add (tmpNode);
    

    
    MobilityHelper mobility;
    mobility.SetMobilityModel ("ns3::IntelligentDriverMobilityModel",
                                  "StartTime", TimeValue (Seconds (curTime)),
                                  "StopTime", TimeValue (endTime),
                                  "MyID", IntegerValue(m_nc.GetN () - 1),
                                  "DesiredTimeHeadway", DoubleValue (timeHeadway),
                                  "DesiredSpeed", DoubleValue (speed),
                                  "DesiredAcceleration", DoubleValue (acc),
                                  "DesiredDeceleration", DoubleValue (dec)
                                  );
    mobility.Install (tmpNode);


    // let user code set application
    m_callback (tmpNode, curTime);



    // application and IDM should be already set by user, now connect
    // other stuff
    
    Ptr<IntelligentDriverMobilityModel> idm = tmpNode->GetObject<IntelligentDriverMobilityModel>();

    Ptr<Navigator> navi = CreateObject<Navigator> (source);
    navi->m_curRoad = source;
    //source->addVehicle(navi);
    navi->m_idm = idm;
    navi->m_id = idm->m_myId;
    idm->m_navi = navi;
  }
}

std::string
ManhattanTopologyHelper::getIntersectionLayout()
{
  std::stringstream ss;
  
  
  ss << "<topology ";
  ss << "type=\"" << "manhattan-grid" << "\" ";
  ss << "blocksh=\"" << m_blocksH << "\" ";
  ss << "blocksv=\"" << m_blocksV << "\" ";
  ss << "blocksizeh=\"" << m_blockSizeH << "\" ";
  ss << "blocksizev=\"" << m_blockSizeV << "\" ";
  ss << "lanewidth=\"" << m_laneWidth << "\" ";
  ss << "sourcelength=\"" << m_sourceLength << "\" ";
  ss << "maindir=\"" << m_mainDir << "\" ";
  ss << "/>" << std::endl;
  
  
  std::vector<Ptr<Intersection> >::iterator iit;
  for (iit = m_intersections.begin() ; iit != m_intersections.end(); ++iit) {
    ss << "<intersection ";
    ss << "id=\"" << (*iit)->m_id << "\" ";
    ss << "posx=\"" << (*iit)->m_pos.x << "\" ";
    ss << "posy=\"" << (*iit)->m_pos.y << "\" ";
    ss << "stopdist=\"" << (*iit)->m_stopDist << "\" ";
    ss << "ibd=\"" << (*iit)->m_ibd << "\" ";
    ss << "/>" << std::endl;
  }
    
  
  std::vector<Ptr<Road> >::iterator rit;
  for (rit = Road::roads.begin() ; rit != Road::roads.end(); ++rit) {
    ss << "<road ";
    ss << "id=\"" << (*rit)->m_id << "\" ";
    
    ss << "startposx=\"" << (*rit)->m_startPos.x << "\" ";
    ss << "startposy=\"" << (*rit)->m_startPos.y << "\" ";
    ss << "length=\"" << (*rit)->m_length << "\" ";
    ss << "dirx=\"" << (*rit)->m_dir.x << "\" ";
    ss << "diry=\"" << (*rit)->m_dir.y << "\" ";
    if ((*rit)->m_start)
      ss << "startid=\"" << (*rit)->m_start->m_id << "\" ";
    if ((*rit)->m_end)
      ss << "endid=\"" << (*rit)->m_end->m_id << "\" ";
    ss << "prio=\"" << (int)(*rit)->m_prio << "\" ";
    ss << "/>" << std::endl;
  }
    
  return ss.str();
}

// i zeile, j spalte
Ptr<Intersection> 
ManhattanTopologyHelper::iget(int i, int j)
{
  uint32_t pos = i * m_blocksH + j;
  //std::cout << pos << " " << i << " " << j << std::endl;
  NS_ASSERT(pos <= m_intersections.size());
  if (pos == m_intersections.size()) {
    NS_LOG_DEBUG ("Creating new Intersection");
    Ptr<Intersection> intersection = CreateObject<Intersection>();
    m_intersections.push_back (intersection);
    return intersection;
  } else {
    return m_intersections[pos];
  }
}

void
ManhattanTopologyHelper::iset(int i, int j, Ptr<Intersection> intersection)
{
NS_ASSERT(false);
}

} // ns3
