/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * 
 *
 */
#include "intelligent-driver-mobility-model.h"
#include "ns3/simulator.h"
#include <math.h>
#include "ns3/node-list.h"
#include "ns3/double.h"
#include "ns3/integer.h"
#include "ns3/log.h"
#include <cassert>

NS_LOG_COMPONENT_DEFINE ("IntelligentDriverMobilityModel");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (IntelligentDriverMobilityModel);


TypeId IntelligentDriverMobilityModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::IntelligentDriverMobilityModel")
    .SetParent<MobilityModel> ()
    .AddConstructor<IntelligentDriverMobilityModel> ()
    .AddAttribute ("StartTime", "The time when the vehicle starts driving",
               TimeValue(Seconds (0.0)),
               MakeTimeAccessor (&IntelligentDriverMobilityModel::m_startTime),
               MakeTimeChecker ())
    .AddAttribute ("StopTime", "The time when the vehicle stops driving",
               TimeValue(Seconds (300.0)),
               MakeTimeAccessor (&IntelligentDriverMobilityModel::m_stopTime),
               MakeTimeChecker ())
    .AddAttribute ("Interval", "Time between movement recalculation",
               TimeValue(Seconds (0.1)),
               MakeTimeAccessor (&IntelligentDriverMobilityModel::m_interval),
               MakeTimeChecker ())
    .AddAttribute ("Direction", "Direction this vehicle is driving to",
               VectorValue(Vector(0, 0, 0)),
               MakeVectorAccessor (&IntelligentDriverMobilityModel::m_dir),
               MakeVectorChecker ())    
    .AddAttribute ("PovID", "ID of heading vehicle",
               IntegerValue (9999999),
               MakeIntegerAccessor (&IntelligentDriverMobilityModel::m_povId),
               MakeIntegerChecker<uint32_t> ())
    .AddAttribute ("MyID", "ID of own vehicle",
               IntegerValue (9999999),
               MakeIntegerAccessor (&IntelligentDriverMobilityModel::m_myId),
               MakeIntegerChecker<uint32_t> ())
    .AddAttribute ("VehicleLength", "Length of vehicles in m",
               DoubleValue (5.0),
               MakeDoubleAccessor (&IntelligentDriverMobilityModel::m_carLength),
               MakeDoubleChecker<double> ())
    .AddAttribute ("DesiredSpeed", "The velocity the vehicle would drive at in free traffic in m/s.",
               DoubleValue (16.0),
               MakeDoubleAccessor (&IntelligentDriverMobilityModel::m_desSpeed),
               MakeDoubleChecker<double> ())
    .AddAttribute ("MinSpace", "A minimum net distance that is kept even at a complete stand-still in a traffic jam in m",
               DoubleValue (2.0),
               MakeDoubleAccessor (&IntelligentDriverMobilityModel::m_minSpace),
               MakeDoubleChecker<double> ())
    .AddAttribute ("DesiredTimeHeadway", "The desired time headway to the vehicle in front in seconds",
               DoubleValue (1.5),
               MakeDoubleAccessor (&IntelligentDriverMobilityModel::m_desHeadway),
               MakeDoubleChecker<double> ())
    .AddAttribute ("DesiredAcceleration", "The desired acceleration in m/s^2",
               DoubleValue (1.0),
               MakeDoubleAccessor (&IntelligentDriverMobilityModel::m_desAcc),
               MakeDoubleChecker<double> ())
    .AddAttribute ("DesiredDeceleration", "The desired deceleration in m/s^2",
               DoubleValue (2.0),
               MakeDoubleAccessor (&IntelligentDriverMobilityModel::m_desDec),
               MakeDoubleChecker<double> ())
    .AddAttribute ("AccLimit", "The maximum deceleration in m/s^2 as a positive value (e.g. 8m/s^2)",
               DoubleValue (8.0),
               MakeDoubleAccessor (&IntelligentDriverMobilityModel::m_accLimit),
               MakeDoubleChecker<double> ())
     .AddTraceSource ("PositionTrace", 
               "The value of the position at each timestep",
                MakeTraceSourceAccessor (&IntelligentDriverMobilityModel::m_positionTrace));
  return tid;
}

IntelligentDriverMobilityModel::IntelligentDriverMobilityModel ()
{
  Simulator::Schedule(m_startTime, &IntelligentDriverMobilityModel::SetStartTimeEvent, this);
  m_speed = 0;
  m_acc = 0;
  m_accExp = 4;
  m_end = true;
  m_stop = false;
  m_povmob = NULL;
  m_astate = FAILSAFE;
}

IntelligentDriverMobilityModel::~IntelligentDriverMobilityModel ()
{
  // if vehicle was on intersection at end of sim, remove from collision list
  // for next run

}

inline Vector
IntelligentDriverMobilityModel::DoGetVelocity (void) const
{
  if (m_end)
    return Vector(0, 0, 0);
  double t = (Simulator::Now () - m_baseTime).GetSeconds ();
  
  double curspeed = m_speed + m_acc*t;
  return Vector (m_dir.x * curspeed,
                 m_dir.y * curspeed,
                 m_dir.z * curspeed);
}

inline Vector
IntelligentDriverMobilityModel::DoGetPosition (void) const
{
  if (m_end)
    return Vector(IDM_INVALID_POS, IDM_INVALID_POS, IDM_INVALID_POS);
  double t = (Simulator::Now () - m_baseTime).GetSeconds ();
  double t_speed = t * m_speed;
  double half_t_square_acc = t*t*0.5*m_acc;
  return Vector (m_pos.x + m_dir.x*t_speed + m_dir.x*half_t_square_acc,
                 m_pos.y + m_dir.y*t_speed + m_dir.y*half_t_square_acc,
                 m_pos.z + m_dir.z*t_speed + m_dir.z*half_t_square_acc);
}

void 
IntelligentDriverMobilityModel::DoSetPosition (const Vector &position)
{
  //std::cout << "set pos to " << position << std::endl; 
  Time now = Simulator::Now ();
  double t = (now - m_baseTime).GetSeconds ();
  m_speed = m_speed + m_acc*t;
  m_baseTime = now;
  m_pos = position;
  NotifyCourseChange ();
}

void
IntelligentDriverMobilityModel::SetDirection (Vector &dir)
{
  double length = sqrt (dir.x*dir.x + dir.y*dir.y);
  if (length != 1.0) {
    m_dir = Vector (dir.x / length, dir.y / length, dir.z / length);
  } else {
    m_dir = dir;
  }
}

void
IntelligentDriverMobilityModel::SetDesiredSpeed (double speed)
{
  m_desSpeed = speed;
}

void
IntelligentDriverMobilityModel::EndMobility (void)
{
  m_end = true;
  Simulator::Cancel (m_nextEvent);
  NS_LOG_DEBUG ("IDM stopped by application at " << Simulator::Now ());
}


void
IntelligentDriverMobilityModel::StopAt (const Vector &position)
{
  NS_LOG_DEBUG (Simulator::Now () << " IDM stopAt " << position);
  m_stoppos = position;
  m_stop = true;
}


void
IntelligentDriverMobilityModel::StopAt (double distance)
{
  if (distance < 0) {
    NS_LOG_UNCOND ("IntelligentDriverMobilityModel::StopAt ignoring negative distance = " << distance);
    return;
  }
  StopAt (Vector (m_pos.x + m_dir.x*distance, 
                  m_pos.y + m_dir.y*distance, 
                  m_pos.z + m_dir.z*distance));
}

bool
IntelligentDriverMobilityModel::GetStop (void)
{
  return m_stop;
}

void
IntelligentDriverMobilityModel::Resume (void)
{
  //NS_LOG_DEBUG (m_stopTime << " IDM resume");
  m_stop = false;
}

void
IntelligentDriverMobilityModel::SetApproach (ApproachState a)
{
  NS_LOG_DEBUG ("IDM " << m_myId << " SetApproach " <<  m_astate 
    << " -> " << a << " navistate= "<< m_navi->m_astate << " at " << Simulator::Now().GetSeconds());
  
  if (m_navi->m_astate == Navigator::CROSSING && m_navi->m_curRoad->m_prio != IDM_HIGH_PRIO) {
    NS_LOG_DEBUG ("IDM " << m_myId << " Can't change state: already crossing" 
    << " at " << Simulator::Now().GetSeconds());
    return;
  } 
  if (m_astate == FAILSAFE && a != FAILSAFE)
    m_navi->EndFailsafe();
  
  
  if (a == RED) {
    m_stop = true;
    StopAt(m_navi->GetStopDist ());
  } else if (a == GREEN) {
    Resume ();
  }
  m_astate = a;
  
}


void
IntelligentDriverMobilityModel::SetStartTimeEvent (void)
{
  if (m_startTime <= Simulator::Now ()) {
    m_baseTime = Simulator::Now ();
    m_end = false;
    m_speed = m_desSpeed;
    
    m_navi->Start();
    m_distDriven = 0;
    m_pos = m_navi->GetCurrentRoadStart ();
    m_dir = m_navi->GetCurrentDirection ();
    m_povmob = m_navi->GetCurrentPOV ();
    
    //std::cout << "IDM starting now " << m_baseTime << std::endl;
    NS_LOG_DEBUG ("IDM " << m_myId << " spawned at " << m_pos << " at " << Simulator::Now().GetSeconds());
    m_nextEvent = Simulator::ScheduleNow (&IntelligentDriverMobilityModel::NextTimestep, this);
  } else {  
    m_nextEvent = Simulator::Schedule(m_startTime, &IntelligentDriverMobilityModel::SetStartTimeEvent, this);
    //std::cout << "IDM starting at " << m_startTime << std::endl;
  }
}

/*
 * called every timestep (e.g. 100ms) to set the current acceleration
 * according to the vehicle's environment.
 * implements the "intelligent driver model".
 * calls UpdatePositionAndSpeed and GetHeadingVehicleSpeedAndDist and
 * schedules next timestep.
 */
void 
IntelligentDriverMobilityModel::NextTimestep (void)
{
  double s;       // s*(v_a, dv_a) in model
  double vdiff;   // speed difference to heading vehicle
  double sdiff;   // distance to heading vehicle
  Navigator::RoadState rs;
  //NS_LOG_LOGIC (" m_speed=" << m_speed << " m_desSpeed=" << m_desSpeed);
  
  UpdatePositionAndSpeed (); // update m_pos and m_speed
  
  // check if we left current road
  rs = m_navi->CheckRoadChange(m_distDriven);
  if (rs == Navigator::CHANGED) {
    m_pos = m_navi->GetCurrentRoadStart ();
    m_dir = m_navi->GetCurrentDirection ();
    m_distDriven = 0;
    m_povmob = m_navi->GetCurrentPOV ();
    NS_LOG_DEBUG ("IDM " << m_myId << " changed road at " << m_pos << " at " << Simulator::Now().GetSeconds());
    
  } else if (rs == Navigator::FINISHED) {
    m_end = true;
    NS_LOG_DEBUG ("IDM " << m_myId << " reached hudson river");
    return;
  }
  
  // how the vehicle approaches an intersection is determined by m_astate
  // the application can set red, green or failsafe (default).
  // 
  if (m_astate == FAILSAFE) {
    m_navi->checkFailsafe(m_speed, 5.0);  // TODO calc with own acc, make configurable
  }
  
  GetHeadingVehicleSpeedAndDist (); // update m_headdist and m_headspeed

  vdiff = m_speed - m_headspeed;
  sdiff = m_headdist;
  
  s = m_minSpace + m_speed * m_desHeadway + (m_speed * vdiff) / (double) (2 * sqrt (m_desAcc * m_desDec));
  
  if (s < m_minSpace)
    s = m_minSpace;

  //std::cout << "s=" << s << " sdiff=" << sdiff << " m_speed=" << m_speed << " m_desSpeed=" << m_desSpeed << std::endl;
  m_acc = m_desAcc * (1 - pow ((m_speed / m_desSpeed), m_accExp) - (s*s)/(sdiff*sdiff));
  if (m_speed  <= 0 &&  m_acc < 0) {
    m_speed = 0;
    m_acc = 0;
  }

  // IDM sometimes uses unrealistic deceleration values, limit to m_accLimit here
  if (m_acc < -1 * m_accLimit) {
    //std::cout << "#### " << Simulator::Now ().GetSeconds() << " " << m_myId << " m_acc=" << m_acc << std::endl;
    m_acc = -1 * m_accLimit;
  }
  
  
  // schedule next timestep
  if (Simulator::Now () + m_interval <= m_stopTime) {
    m_nextEvent = Simulator::Schedule(m_interval, &IntelligentDriverMobilityModel::NextTimestep, this);
  } else {
    NS_LOG_DEBUG ("IDM reached stoptime at " << m_stopTime << " m_myId=" << m_myId);
  }
  
  NS_LOG_INFO ("time=" << Simulator::Now().GetSeconds () <<  " Id="<< m_myId <<" pos=" << m_pos << " speed=" << m_speed << " acc=" << m_acc);
  m_positionTrace(Simulator::Now (), m_myId, m_pos, m_dir, m_speed);
}

/*
 * updates vehicles position and speed according to the acceleration.
 * this is the only function, that should modify m_pos, m_speed and 
 * m_baseTime except for functions explicitly setting the position
 * (i.e. DoSetPosition)
 */
void
IntelligentDriverMobilityModel::UpdatePositionAndSpeed (void)
{
  Time now = Simulator::Now ();
  double t = (now - m_baseTime).GetSeconds ();
  if (t > 0) {
    double t_speed = t * m_speed;
    double half_t_square_acc = t*t*0.5*m_acc;
    double t_speed_half_t_square_acc = t_speed + half_t_square_acc;
    
    m_pos =  Vector (m_pos.x + m_dir.x*t_speed_half_t_square_acc,
                     m_pos.y + m_dir.y*t_speed_half_t_square_acc,
                     m_pos.z + m_dir.z*t_speed_half_t_square_acc);
    m_distDriven += t_speed_half_t_square_acc;
    //std::cout << "t=" << t <<std::endl; 
    m_speed = m_speed + m_acc*t;
  }
  m_baseTime = now;
  
  
}

double IntelligentDriverMobilityModel::GetCarLength (void)
{
  return m_carLength;
}

/*
 * updates distance and speed of heading vehicle.
 * if a stop position is given, the closer will be choosen
 * if no heading vehicle exists and no stop position has been set,
 * distance will be set to 99999 and relative speed to 0
 */
void
IntelligentDriverMobilityModel::GetHeadingVehicleSpeedAndDist (void)
{
  double vdist = -1;   // distance to heading vehicle (if any)
  double sdist = -1;   // distance to stop point
  Vector povSpeedV;
  Vector povPos;

  // we have a heading vehicle
  if (m_povmob) {
    povPos = m_povmob->GetPosition ();
    povSpeedV = m_povmob->GetVelocity ();
    vdist = CalculateDistance (m_pos, povPos) - m_carLength;
    /*std::cout << "####++++ " << Simulator::Now ().GetSeconds() << " " << m_myId 
      << " mypos " << m_pos 
      << " povpos " << povPos << std::endl;
    */
    // ensure we are behind pov
    if (((povPos.x - m_pos.x) * m_dir.x < 0 || (povPos.y - m_pos.y) * m_dir.y < 0 )
        && povPos.x != IDM_INVALID_POS) {
      // we are moving away from POV, this should not happen!
      // set stop  point in order to wait till POV proceeds
      // overwrite speed here
      m_speed = 0;
      m_headspeed = 0;
      m_headdist = m_minSpace;
      //std::cout << "####++++ " << Simulator::Now ().GetSeconds() << " " << m_myId << " IDM WAIT at " << m_pos << std::endl;
      return;
    }

  }
  
  
  // we have a stop point
  if (m_stop) {
    sdist = CalculateDistance (m_pos, m_stoppos) - (double)(m_carLength/2);
  }
  
  if (m_povmob && m_stop) { // heading vehicle and stop point exist
    m_headspeed = sqrt (povSpeedV.x*povSpeedV.x+povSpeedV.y*povSpeedV.y);
    double speeddiff = m_speed - m_headspeed;
    if (m_speed <= 0) {
      // i am not moving: whatever is closer
      if (sdist < vdist) {
        m_headspeed = 0;
        m_headdist = sdist;
      } else {
        m_headspeed = sqrt (povSpeedV.x*povSpeedV.x+povSpeedV.y*povSpeedV.y);
        m_headdist = vdist;      
      }
    } else if (speeddiff <= 0) {
      // i am moving but other vehicle is not coming closer: stoppoint
      m_headspeed = 0;
      m_headdist = sdist;
      
    } else {  
      // i am moving and other vehicle is coming closer
      // speeddiff > 0 => m_speed > 0
      if (sdist / m_speed <= vdist / speeddiff) {
        m_headspeed = 0;
        m_headdist = sdist;
      } else {
        m_headspeed = sqrt (povSpeedV.x*povSpeedV.x+povSpeedV.y*povSpeedV.y);
        m_headdist = vdist;
      }
    }
    
    
  } else if (m_povmob) { // only heading vehicle
    m_headspeed = sqrt (povSpeedV.x*povSpeedV.x+povSpeedV.y*povSpeedV.y);
    m_headdist = vdist;
  } else if (m_stop) { // only stop point
    m_headspeed = 0;
    m_headdist = sdist;
  } else {    // free road
    m_headspeed = 0;
    m_headdist = 999999;
  }
    
}

double
IntelligentDriverMobilityModel::GetSpeed (void)
{
  return m_speed;
}

double
IntelligentDriverMobilityModel::GetAcc (void)
{
  return m_acc;
}

} // namespace ns3
