/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */


#include "propagation-loss-model.h"
#include "virtualsource11p-propagation-loss-model.h"
#include "ns3/log.h"
#include "ns3/mobility-model.h"
#include "ns3/boolean.h"
#include "ns3/double.h"
#include <math.h>
#include "ns3/object-factory.h"
#include "ns3/random-variable.h"
#include "ns3/simulator.h"
#include "virtualsource11p-propagation-loss-model.h"


namespace ns3 {


NS_LOG_COMPONENT_DEFINE ("VirtualSource11pPropagationLossModel");
NS_OBJECT_ENSURE_REGISTERED (VirtualSource11pPropagationLossModel);

TypeId 
VirtualSource11pPropagationLossModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::VirtualSource11pPropagationLossModel")
    .SetParent<PropagationLossModel> ()
    .AddConstructor<VirtualSource11pPropagationLossModel> ()
    .AddAttribute ("RoadSize", "The road's and crossing's width (in m) TODO half.", //TODO half
                   DoubleValue (10.0),
                   MakeDoubleAccessor (&VirtualSource11pPropagationLossModel::m_roadsize),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("Lambda",
                   "The wavelength  (default is 5.9 GHz at 300 000 km/s).",
                   DoubleValue (300000000.0 / 5.90e9),
                   MakeDoubleAccessor (&VirtualSource11pPropagationLossModel::m_lambda),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("SystemLoss", "The system loss",
                   DoubleValue (4.8),
                   MakeDoubleAccessor (&VirtualSource11pPropagationLossModel::m_systemLoss),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("HeightAboveZ",
                   "The height of the antenna (m) above the node's Z coordinate",
                   DoubleValue (1.5),
                   MakeDoubleAccessor (&VirtualSource11pPropagationLossModel::m_heightAboveZ),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("CrossingCenter",
                   "The coordinates of the crossing's center.",
                   VectorValue (Vector (0, 0, 0)),
                   MakeVectorAccessor (&VirtualSource11pPropagationLossModel::m_crossingCenter),
                   MakeVectorChecker ())
    .AddAttribute ("IsSuburban",
                   "Environment is suburban.",
                   BooleanValue (true),
                   MakeBooleanAccessor (&VirtualSource11pPropagationLossModel::m_isSuburban),
                   MakeBooleanChecker ())
  ;
  return tid;
}
VirtualSource11pPropagationLossModel::VirtualSource11pPropagationLossModel ()
  : PropagationLossModel ()
{

  m_erlangRandomVariable = CreateObject<ErlangRandomVariable> ();
  
  m_gammaRandomVariable = CreateObject<GammaRandomVariable> ();
  
  m_normVar = CreateObject<NormalRandomVariable> ();
  m_normVar->SetAttribute ("Mean", DoubleValue (0.0));
  m_normVar->SetAttribute ("Variance", DoubleValue (4.1*4.1));

  m_threeLogDistPLM = CreateObject<ThreeLogDistancePropagationLossModel> ();
}

VirtualSource11pPropagationLossModel::~VirtualSource11pPropagationLossModel ()
{
}

void
VirtualSource11pPropagationLossModel::SetRoadSize (double rs)
{
  m_roadsize = rs;
}

double
VirtualSource11pPropagationLossModel::DoCalcRxPower (double txPowerDbm,
                                  Ptr<MobilityModel> a,
                                  Ptr<MobilityModel> b) const
{
  return DoCalcRxPower (txPowerDbm, a, b, true);
}

double
VirtualSource11pPropagationLossModel::DoCalcRxPowerNoFad (double txPowerDbm,
                                  Ptr<MobilityModel> a,
                                  Ptr<MobilityModel> b) const
{
  return DoCalcRxPower (txPowerDbm, a, b, false);
}

bool
VirtualSource11pPropagationLossModel::GetCommonIntersection (Ptr<MobilityModel> a,
                      Ptr<MobilityModel> b, Vector &cs, double &rs, bool & suburban) const
{

  Ptr<IntelligentDriverMobilityModel> idm1 = DynamicCast<IntelligentDriverMobilityModel> (a);
  Ptr<IntelligentDriverMobilityModel> idm2 = DynamicCast<IntelligentDriverMobilityModel> (b);
  Ptr<Intersection> intersection;
  intersection = idm1->m_navi->GetCommonIntersection(idm2->m_navi);
  if (intersection != NULL) {
    cs = intersection->m_pos;
    rs = intersection->m_ibd/2.0;
    suburban = intersection->m_suburban;
    return true;
  } else {
    return false;
  }
}

double
VirtualSource11pPropagationLossModel::DoCalcRxPower (double txPowerDbm,
                                  Ptr<MobilityModel> a,
                                  Ptr<MobilityModel> b,
                                  bool withfading) const
{
  Vector srcPos, dstPos;
  double nlosPowerDbm, rxPowerDbm, rxPowerDbmFad;
  double dr, dt;
  double losPrDbm;
  double Pr0;
  double xt;
  Vector crossingCenter = m_crossingCenter;
  double roadsize = m_roadsize;
  bool isSuburban = m_isSuburban;
  bool los = false;
  
  if (!GetCommonIntersection (a, b, crossingCenter, roadsize, isSuburban)) {
    // no common intersection -> NLOS2 -> assume no signal -> return -99999
    //NS_LOG_DEBUG (Simulator::Now().GetSeconds() << " VirtualSource11pPropagationLossModel: srcPos=" << a->GetPosition () << " dstPos=" <<
    //b->GetPosition () << " No common intersection");
    return -99999999;
  } 
  
  // vehicle positions are shifted so that crossing center is virtually
  // at (0, 0) for easier processing
  srcPos = vdiff(a->GetPosition (), crossingCenter);
  dstPos = vdiff(b->GetPosition (), crossingCenter);

  
  // check if vehicles are in LOS
  // this checks if both vehicles are on the same road
  // or if sender or receiver is on intersection
  if (srcPos.x <= roadsize && srcPos.x >= -1 * roadsize
    && dstPos.x <= roadsize && dstPos.x >= -1 * roadsize)
    los = true;
  if (srcPos.y <= roadsize && srcPos.y >= -1 * roadsize
    && dstPos.y <= roadsize && dstPos.y >= -1 * roadsize)
    los = true;
  

  if (!los) { 
    // vehicles are not on same road
    // check if they are LOS on perpendicular roads

    // distances to virtualSource = crossingCenter
    dr = std::sqrt(dstPos.x*dstPos.x + dstPos.y*dstPos.y);
    dt = std::sqrt(srcPos.x*srcPos.x + srcPos.y*srcPos.y);
    
    // sender distance to wall
    if (srcPos.y >= -1 * roadsize && srcPos.y <= roadsize) {  // sender is on road y=~0
      xt = (dstPos.y >= 0) ? roadsize - srcPos.y : srcPos.y + roadsize;
    } else {
      xt = (dstPos.x >= 0) ? roadsize - srcPos.x : srcPos.x + roadsize;
    }

    // check if vehicles are in LOS
    // this check assumes that both vehicles are on perpendicular roads
    double losDr = 0; // the receiver distance to virtual source where NLOS->LOS
    losDr = dt * (xt / (dt - roadsize));  // TODO: this calculation might not be 100% correct
        // as dt is distance to center and not parallel to wall
    if (dr <= losDr) {
      los = true;
    }
  }

  
  if (!los) {
    // NLOS rx power calculation according to formula (p. 78)
    // done in dBm space!
    Pr0 = getPr0 (dr, dt, xt, roadsize, isSuburban);
    nlosPowerDbm = txPowerDbm - m_systemLoss - Pr0;
  
    // model NLOS fading with a normal distribution with sigma = 4.1dB   
    rxPowerDbm = nlosPowerDbm;
    rxPowerDbmFad = nlosPowerDbm + m_normVar->GetValue();
    
  } else {
      // simply use LOS, no combination of LOS and NLOS!
      
    double plRef = 47.86; // reference loss at 1m
    Vector vdist = vdiff(srcPos, dstPos);
    double dist = std::sqrt(vdist.x*vdist.x + vdist.y*vdist.y);
    
    if (dist > 1) {
      losPrDbm = txPowerDbm - m_systemLoss - plRef - 10 * 2.04 * log10(dist);
    } else {
      losPrDbm = txPowerDbm - m_systemLoss - plRef;
    }
    rxPowerDbm = losPrDbm;
    
    // model LOS fading with a NAK-1 distribution with std-dev = 4.1dB  
    rxPowerDbmFad = NakagamiCalcRxPower (rxPowerDbm, a, b);
  }
  
  /*NS_LOG_DEBUG (Simulator::Now().GetSeconds() << " VirtualSource11pPropagationLossModel: srcPos=" << a->GetPosition () << " dstPos=" <<
    b->GetPosition () << " common intersection at " << crossingCenter <<" LOS=" << los <<" rxPowerDbm=" << rxPowerDbm << 
    " rxPowerDbmFad=" << rxPowerDbmFad);
  */
  //double antennaGain = 3.0;
  //std::cout << ((dstPos.y < 0) ? -1*dr : dr )<< " " << rxPowerDbmFad+antennaGain << std::endl;
  //std::cout << ((dstPos.y < 0) ? -1*dr : dr )<< " " << nlosPowerDbm  << " " << losPrDbm << " " << rxPowerDbm << std::endl;
  
  return withfading ? rxPowerDbmFad : rxPowerDbm;
}

/* NLOS rx power calculation according to formula (p. 78)
 * done in dBm space!
 */
double
VirtualSource11pPropagationLossModel::getPr0 (double dr, double dt, double xt, double roadsize, bool isSuburban) const
{
  const double PI = 3.14159265358979323846;
  double d_b = 4*m_heightAboveZ*m_heightAboveZ / m_lambda;
  float exp_dt = 0.957;
  float exp_st = 0.81;
  double suburban_w = isSuburban ? 2.94 : 0.0;
  double Pr0_t = 0.0;
  
  if (dr <= d_b) {
    Pr0_t = pow(dt,exp_dt) / ((double)pow(xt * 2 * roadsize,exp_st)) *
      (4*PI*dr)/((double)m_lambda);
  } else {
    Pr0_t = pow(dt,exp_dt) / ((double)pow(xt * 2 * roadsize,exp_st)) *
      (4*PI*dr*dr)/((double)m_lambda*d_b);    // dr^2 and lambda*d_b if dr > db
  }
  
  return 3.75 + suburban_w + 10*log10 ((double)pow(Pr0_t, 2.69));
}

Vector
VirtualSource11pPropagationLossModel::vplus (const Vector v, const Vector w) const
{
  return Vector (v.x + w.x, v.y + w.y, v.z + w.z);
}

Vector
VirtualSource11pPropagationLossModel::vdiff (const Vector v, const Vector w) const
{
  return Vector (v.x - w.x, v.y - w.y, v.z - w.z);
}

double
VirtualSource11pPropagationLossModel::NakagamiCalcRxPower (double txPowerDbm,
                                             Ptr<MobilityModel> a,
                                             Ptr<MobilityModel> b) const
{

  double distance = a->GetDistanceFrom (b);
  NS_ASSERT (distance >= 0);

  double m = 1;

  // the current power unit is dBm, but Watt is put into the Nakagami /
  // Rayleigh distribution.
  double powerW = pow (10, (txPowerDbm - 30) / 10);

  double resultPowerW;

  // switch between Erlang- and Gamma distributions: this is only for
  // speed. (Gamma is equal to Erlang for any positive integer m.)
  unsigned int int_m = static_cast<unsigned int>(floor (m));

  if (int_m == m)
    {
      resultPowerW = m_erlangRandomVariable->GetValue (int_m, powerW / m);
    }
  else
    {
      resultPowerW = m_gammaRandomVariable->GetValue (m, powerW / m);
    }

  double resultPowerDbm = 10 * log10 (resultPowerW) + 30;

  NS_LOG_DEBUG ("Nakagami distance=" << distance << "m, " <<
                "power=" << powerW <<"W, " <<
                "resultPower=" << resultPowerW << "W=" << resultPowerDbm << "dBm");

  return resultPowerDbm;
}


int64_t
VirtualSource11pPropagationLossModel::DoAssignStreams (int64_t stream)
{

  m_normVar->SetStream (stream);
  m_erlangRandomVariable->SetStream (stream+1);
  m_gammaRandomVariable->SetStream (stream+2);
  return 3;
}



// ------------------------------------------------------------------------- //

} // namespace ns3

